@@ -1,9 +1,8 @@
 import os
+import socket
 from abc import ABC
 from datetime import datetime
-from http.client import HTTPConnection
 from pathlib import Path
-from socket import socket
 from ssl import SSLSocket
 from typing import Union, Optional, Dict, List, Generator, Tuple
 
@@ -222,6 +221,7 @@ class CacheAdapter(HTTPAdapter, ABC):
 
     def _handle_response(self, response: Response):
         if self._is_passive:  # Store received response in cache
+            self._prepare_response(response)
             self._store_response(response)
             logger.debug(f"{self._indent}  Saved response in cache")
         self._post_send(response)
@@ -230,10 +230,7 @@ class CacheAdapter(HTTPAdapter, ABC):
         self._timestamp = datetime.now().astimezone()
 
     def _post_send(self, response: Response):
-        conn = getattr(response.raw, '_connection')
-        if conn is not None and not getattr(conn.sock, '_closed'):
-            conn.sock.close()
-            setattr(response.raw, '_connection', None)
+        pass
 
     def _send(self, request: PreparedRequest, stream=False, timeout=None, verify=True,
               cert=None, proxies=None) -> Response:
@@ -262,10 +259,18 @@ class CacheAdapter(HTTPAdapter, ABC):
             f"Could not find cached request for [{request.method}] {request.url}")
 
     def _store_response(self, response: Response):
+        raise NotImplementedError()
+
+    def _prepare_response(self, response: Response):
         setattr(response, 'timestamp', self._timestamp)
-        conn: HTTPConnection = getattr(response.raw, '_connection', None)
-        if conn is not None and not getattr(conn.sock, '_closed'):
-            sock: socket = conn.sock
+        sock: socket
+        try:
+            sock = socket.fromfd(response.raw.fileno(), socket.AF_INET, socket.SOCK_STREAM)
+        except IOError as e:
+            logger.debug(f"No response socket: {e}")
+            return
+        setattr(response.raw, 'sock', sock)
+        if sock is not None and not getattr(sock, '_closed'):
             if isinstance(sock, SSLSocket):
                 # Save ssl certificate info
                 cert: dict = sock.getpeercert()
@@ -278,8 +283,7 @@ class CacheAdapter(HTTPAdapter, ABC):
             # Save connection info
             setattr(response, 'ip', ip)
             setattr(response, 'port', port)
-            conn.sock.close()
-            setattr(response.raw, '_connection', None)
+            sock.close()
 
     def _delete(self, response: Response, **kwargs):
         raise NotImplementedError()
@@ -456,7 +460,6 @@ class MemoryCacheAdapter(CacheAdapter, ABC):
         return False
 
     def _store_response(self, response: Response):
-        super(MemoryCacheAdapter, self)._store_response(response)
         url = parse_url(response.request.url)
         origin = self._entries.get(url.hostname, {})
         responses = origin.get(url.path, [])
